@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * wsdl-thread.h: Thread together all the pieces of the WSDL document
  *
@@ -12,6 +13,6 @@
 
 #include "wsdl-parse.h"
 
-extern int wsdl_thread(wsdl_definitions *);
+int wsdl_thread (wsdl_definitions *definitions);
 
 #endif /* _WSDL_THREAD_H_ */
